from sunyata.http.factory import HttpFactory
from sunyata.http.status import *
import uvicorn
from sunyata.http.rawserver import RawHttpServer
from sunyata.http.request import HttpRequest
import sunyata.util as util
import traceback


class HttpServer(RawHttpServer):

    routerMap = {}

    def __init__(self, bind = '0.0.0.0', port=9989, log_level='error'):
        RawHttpServer.__init__(self, bind=bind, port=port)
        self.config = uvicorn.Config("sunyata.http.server:HttpServer", host=self.bind, port=self.port, log_level=log_level)

    def serve(self):
        server = uvicorn.Server(self.config)
        print('http running on http://%s:%s' % (self.bind, self.port) )
        server.run()

    async def asyncServe(self):
        server = uvicorn.Server(self.config)
        server.config.setup_event_loop()
        await server.serve()
    
    async def genHttpRequest(self, scheme, httpVersion, method, path, queryString, headers, body, clientTuple):
        httpRequest = HttpRequest()
        httpRequest.scheme = scheme
        httpRequest.httpVersion = httpVersion
        httpRequest.method = method
        httpRequest.uri = path
        httpRequest.queryString = queryString
        httpRequest.headers = headers
        httpRequest.body = body
        httpRequest.client.remoteAddr = clientTuple[0]
        httpRequest.client.port = clientTuple[1]
        httpRequest.data = HttpFactory.parseRequestData(util.str2bytes(path + "?") + queryString, body)
        return httpRequest

    async def readBody(self, receive):
        """
        Read and return the entire body from an incoming ASGI message.
        """
        body = b''
        more_body = True
        while more_body:
            message = await receive()
            body += message.get('body', b'')
            more_body = message.get('more_body', False)
        return body
    
    async def rawHeadersToDict(self, rawHeaders):
        headers = {}
        for line in rawHeaders:
            headers[util.bytes2str(line[0])] = str(util.bytes2str(line[1]))
        return headers

    async def __call__(self, scope, receive, send):
        assert scope['type'] == 'http'
        headers = await self.rawHeadersToDict(scope['headers'])
        body = await self.readBody(receive)
        try:
            httpRequest = await self.genHttpRequest(
                scope['scheme'],
                scope['http_version'],
                scope['method'], 
                scope['path'],  
                scope['query_string'],
                headers, 
                body,
                scope['client']
            )
            httpResponse = await self.handleRequest(httpRequest)
        except Exception as ex:
            print(str(ex) + ' ' + traceback.format_exc())
            httpResponse = HttpFactory.genHttpResponse(HttpStatus500, str(ex))
        await send(
            {
                "type": "http.response.start",
                "status": httpResponse.status.code,
                "headers": [[b"content-type", b"text/plain"]],
            }
        )
        await send(
            {
                "type": "http.response.body", 
                "body": httpResponse.responseBody(),
            }
        )


route = HttpServer.route


import operator
from sunyata.eventloop import EventLoop

class RpcMethod(object):

    __slots__ = ('_methodType', '_method', '_classDefine', '_isCoroutine', '_methodType')

    TYPE_WITHOUT_CLASS = 0
    TYPE_WITH_CLASS = 1

    def __init__(self, methodType, method, classDefine = None, isCoroutine = False):
        self._methodType = methodType
        self._method = method
        self._classDefine = classDefine
        self._isCoroutine = isCoroutine
        if self._methodType == self.TYPE_WITH_CLASS and self._classDefine == None:
            raise Exception('classDefine is None')

    def call(self, *args, **kwargs):
        if self._methodType == self.TYPE_WITH_CLASS:
            classInstance = self._classDefine()
            if self._isCoroutine:
                resp = EventLoop.runUntilComplete( operator.methodcaller(self._method.__name__, *args, **kwargs)(classInstance) )
            else:
                resp = operator.methodcaller(self._method.__name__, *args, **kwargs)(classInstance)
        else:
            if self._isCoroutine:
                resp = EventLoop.runUntilComplete( self._method(*args, **kwargs) )
            else:
                resp = self._method(*args, **kwargs)
        return resp

    async def asyncCall(self, *args, **kwargs):
        if self._methodType == self.TYPE_WITH_CLASS:
            classInstance = self._classDefine()
            if self._isCoroutine:
                resp = await ( operator.methodcaller(self._method.__name__, *args, **kwargs)(classInstance) )
            else:
                resp = operator.methodcaller(self._method.__name__, *args, **kwargs)(classInstance)
        else:
            if self._isCoroutine:
                resp = await self._method(*args, **kwargs)
            else:
                resp = self._method(*args, **kwargs)
        return resp


from sunyata.rpc.protocal import TcpProtocal, UdpProtocal, HttpProtocal, RpcProtocal
import multiprocessing
from sunyata.rpc.exception import FuncNotFoundException
import queue
import threading
import socket
from sunyata.rpc.discovery import DiscoveryConfig, RpcDiscovery
import struct
from sunyata.rpc.compress import RpcCompress
from types import FunctionType
from sunyata.rpc.method import RpcMethod
import asyncio
import uvloop
import inspect
from sunyata.http.server import HttpServer
asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


class RpcServer(object):

    funcMap = {}
    funcList = []

    def __init__(self):
        self.discoveryConfig = None
        self.discovery = None
        self.protocal = RpcProtocal()

    @classmethod
    def regist(cls, func):
        if isinstance(func, FunctionType):
            if inspect.iscoroutinefunction(func):
                cls.funcMap[ func.__name__ ] = RpcMethod(RpcMethod.TYPE_WITHOUT_CLASS, func, isCoroutine=True)
                cls.funcList = cls.funcMap.keys()
            else:
                cls.funcMap[ func.__name__ ] = RpcMethod(RpcMethod.TYPE_WITHOUT_CLASS, func)
                cls.funcList = cls.funcMap.keys()
        else:
            classDefine = func
            serMethods = list( filter(lambda m: not m.startswith('_'), dir(classDefine)) )
            for methodName in serMethods:
                funcName = "{}.{}".format(classDefine.__name__, methodName)
                funcObj = getattr(classDefine, methodName)
                if inspect.iscoroutinefunction(funcObj):
                    cls.funcMap [ funcName ] = RpcMethod(RpcMethod.TYPE_WITH_CLASS, funcObj, classDefine, isCoroutine=True)
                    cls.funcList = cls.funcMap.keys()
                else:
                    cls.funcMap [ funcName ] = RpcMethod(RpcMethod.TYPE_WITH_CLASS, funcObj, classDefine)
                    cls.funcList = cls.funcMap.keys()

    @classmethod
    def getRegistedMethods(cls):
        return list(cls.funcMap.keys())

    def run(self, func, args, kwargs):
        try:
            if func not in self.funcList:
                return FuncNotFoundException('func not found')
            methodObj = self.funcMap[func]
            args = tuple(args)
            if len(args) == 0 and len(kwargs) == 0:
                resp = methodObj.call()
            else:
                resp = methodObj.call(*args, **kwargs)
            return resp
        except Exception as ex:
            return Exception('server exception, ' + str(ex))

    def setDiscoverConfig(self, config: DiscoveryConfig):
        self.discoveryConfig = config
        self.discovery = RpcDiscovery(self.discoveryConfig.consulHost, self.discoveryConfig.consulPort, self.discoveryConfig.consulToken, self.discoveryConfig.etcdHost, self.discoveryConfig.etcdPort)

    def setKeepaliveTimeout(self, keepaliveTimeout: int):
        self.protocal.transport.setKeepaliveTimeout(keepaliveTimeout)

    @classmethod
    def rpc(cls, func):
        cls.regist(func)
        return func
    
    def printLogo(self):
        logo = """"""
        print(logo)


class SimpleTcpRpcServer(RpcServer):
    
    def __init__(self, host, port):
        RpcServer.__init__(self)
        self.host = host
        self.port = port
        self.protocal = TcpProtocal(host, port)
    
    def serve(self):
        self.protocal.transport.bind(self.keepaliveTimeout)
        self.printLogo()
        while 1:
            msg = self.protocal.transport.recv()
            request = self.protocal.unserialize(msg)
            func, args, kwargs = self.protocal.parseRequest(request)
            resp = self.run(func, args, kwargs)
            self.protocal.transport.send(self.protocal.serialize(resp))


class BlockTcpRpcServer(SimpleTcpRpcServer):
    
    def __init__(self, host, port, workers = multiprocessing.cpu_count()):
        SimpleTcpRpcServer.__init__(self, host, port)
        self.worker = workers

    def handle(self, conn):
        while 1:
            try:
                msg = self.protocal.transport.recvPeer(conn)
                request = self.protocal.unserialize(msg)
                func, args, kwargs = self.protocal.parseRequest(request)
                resp = self.run(func, args, kwargs)                    
                self.protocal.transport.sendPeer(self.protocal.serialize(resp), conn)
            except Exception as ex:
                conn.close()
                return

    def serve(self):
        self.protocal.transport.bind()
        self.printLogo()
        if self.discovery and self.discoveryConfig:
            self.discovery.regist(self.discoveryConfig.serviceName, self.discoveryConfig.serviceHost, self.discoveryConfig.servicePort, ttlHeartBeat=True)
        while 1:
            conn, _ = self.protocal.transport.accept()
            t = threading.Thread(target=self.handle, args=(conn,) )
            t.setDaemon(True)
            t.start()


class HttpRpcServer(RpcServer):
    
    def __init__(self, host, port, workers = multiprocessing.cpu_count()):
        RpcServer.__init__(self)
        self.host = host
        self.port = port
        self.protocal = HttpProtocal(host, port, workers)
        self.app = HttpServer(self.host, self.port)
        self.app.addRoute('/', self.handle)

    def handle(self, request):
        body = request.body
        resp = self.callback(body)
        return resp
        
    def callback(self, package):
        isEnableCompress = package[:1]
        msg = package[1:]
        if isEnableCompress == b'1':
            msg = RpcCompress.decompress(msg)
        request = self.protocal.unserialize(msg)
        func, args, kwargs = self.protocal.parseRequest(request)
        resp = self.run(func, args, kwargs)
        resp = self.protocal.serialize(resp)
        return resp
    
    async def asyncServe(self):
        tasks = []
        tasks.append(self.app.asyncServe())
        if self.discovery and self.discoveryConfig:
            registTask = asyncio.create_task(self.discovery.asyncRegist(self.discoveryConfig.serviceName, self.discoveryConfig.serviceHost, self.discoveryConfig.servicePort, ttlHeartBeat=True))
            tasks.append(registTask)
        await asyncio.wait(tasks)

    def serve(self):
        asyncio.run(self.asyncServe())
        """
        tRegist = None
        if self.discovery and self.discoveryConfig:
            self.discovery.regist(self.discoveryConfig.serviceName, self.discoveryConfig.serviceHost, self.discoveryConfig.servicePort, ttlHeartBeat=True)
        self.printLogo()
        print('http rpc running on http://%s:%s' % (self.host, self.port) )
        self.app.serve()
        """


class TcpRpcServer(BlockTcpRpcServer):

    def __init__(self, host, port):
        BlockTcpRpcServer.__init__(self, host, port)
        self.host = host
        self.port = port
    
    async def handle(self, reader, writer):
        while 1:
            try:
                data = await reader.read(5)
                if data == b'':
                    raise Exception('peer closed')
                lengthField = data[:4]
                compressField = data[4:5]
                isEnableCompress = 0
                if compressField == b'1':
                    isEnableCompress = 1
                toread = struct.unpack("i", lengthField)[0]
                msg = b''
                readn = toread 
                while 1:
                    rmsg = await reader.read(readn)
                    if rmsg == b'':
                        raise Exception('peer closed')
                    msg = msg + rmsg
                    if len(msg) == toread:
                        readn = toread - len(rmsg)
                        break
                if isEnableCompress:
                    msg = RpcCompress.decompress(msg)
                request = self.protocal.unserialize(msg)
                func, args, kwargs = self.protocal.parseRequest(request)
                resp = await self.run(func, args, kwargs) 
                respbytes = self.protocal.serialize(resp)

                isEnableCompress = b'0'
                if len(msg) >= RpcCompress.enableCompressLen:
                    isEnableCompress = b'1'
                    respbytes = RpcCompress.compress(respbytes)
                lenbytes = struct.pack("i", len(respbytes))
                writer.write( lenbytes + isEnableCompress + respbytes)
            except Exception as ex:
                #print(ex, traceback.format_exc())
                writer.close()
                return

    async def main(self):
        loop = asyncio.get_event_loop()
        coro = asyncio.start_server(self.handle, self.host, self.port, loop=loop)
        server = loop.run_until_complete(coro)
        print('tcp rpc running on tcp://%s:%s' % (self.host, self.port))
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            pass
        # Close the server
        server.close()
        loop.run_until_complete(server.wait_closed())
        loop.close()

    async def run(self, func, args, kwargs):
        try:
            if func not in self.funcList:
                return FuncNotFoundException('func not found')
            methodObj = self.funcMap[func]
            args = tuple(args)
            if len(args) == 0 and len(kwargs) == 0:
                resp = await methodObj.asyncCall()
            else:
                resp = await methodObj.asyncCall(*args, **kwargs)
            return resp
        except Exception as ex:
            return Exception('server exception, ' + str(ex))

    def serve(self):
        loop = asyncio.get_event_loop()
        tasks = []
        tRegist = None
        if self.discovery and self.discoveryConfig:
            tRegist = self.discovery.asyncRegist(self.discoveryConfig.serviceName, self.discoveryConfig.serviceHost, self.discoveryConfig.servicePort, ttlHeartBeat=True)
            tasks.append(tRegist)
        coro = asyncio.start_server(self.handle, self.host, self.port, loop=loop)
        tasks.append(coro)
        rs = loop.run_until_complete(asyncio.gather(*tasks))
        self.printLogo()
        print('tcp rpc running on tcp://%s:%s' % (self.host, self.port) )
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            pass
        server = rs[-1]
        # Close the server
        server.close()
        loop.run_until_complete(server.wait_closed())
        loop.close()


class UdpRpcServer(RpcServer):

    def __init__(self, host, port, workers = multiprocessing.cpu_count()):
        RpcServer.__init__(self)
        self.protocal = UdpProtocal(host, port)
        self.worker = workers
        self.queueMaxSize = 100000
        self.queue = queue.Queue(self.queueMaxSize)
        self.host = host
        self.port = port

    def startWorkers(self):
        for i in range(self.worker):
            t = threading.Thread(target=self.handle)
            t.setDaemon(True)
            t.start()

    def handle(self):
        while 1:
            try:
                body = self.queue.get()
                addr = body.get('addr')
                msg = body.get('msg')
                request = self.protocal.unserialize(msg)
                func, args, kwargs = self.protocal.parseRequest(request)
                resp = self.run(func, args, kwargs)
                self.protocal.transport.sendPeer(self.protocal.serialize(resp), addr = addr)
            except Exception as ex:
                print('UDP handler exception:', ex)
    
    def serve(self):
        self.startWorkers()
        self.protocal.transport.bind()
        self.printLogo()
        print('udp rpc running on udp://%s:%s' % (self.host, self.port))
        if self.discovery and self.discoveryConfig:
            self.discovery.regist(self.discoveryConfig.serviceName, self.discoveryConfig.serviceHost, self.discoveryConfig.servicePort, ttlHeartBeat=True)
        while 1:
            try:
                msg, cliAddr = self.protocal.transport.recv()
                self.queue.put({'msg' : msg, 'addr' : cliAddr})
            except socket.timeout:
                pass


rpc = RpcServer.rpc


from socket import *

class TcpTransport(object):

    __slots__ = ('host', 'port', 'timeout', 'socket', 'keepaliveTimeout', 'backlog')

    def __init__(self, host, port, timeout = 30, keepaliveTimeout=30):
        self.host = host
        self.port = port
        self.timeout = timeout
        self.socket = None
        self.keepaliveTimeout = keepaliveTimeout
        self.backlog = 100

    def bind(self):
        self.socket = socket(AF_INET, SOCK_STREAM)
        self.socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self.socket.bind( (self.host, self.port) )
        self.socket.listen(self.backlog)

    def setKeepaliveTimeout(self, keepaliveTimeout: int):
        self.keepaliveTimeout = keepaliveTimeout

    def connect(self):
        if self.socket == None:
            self.socket = socket(AF_INET, SOCK_STREAM)
            self.socket.settimeout(self.timeout)
            self.socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
            self.socket.connect( (self.host, self.port) )

    def accept(self):
        cli, addr = self.socket.accept()
        cli.settimeout(self.keepaliveTimeout)
        return cli, addr

    def close(self):
        if self.socket:
            self.socket.close()
            self.socket = None

    def reconnect(self):
        self.close()
        self.connect()
        print('tcp client reconnect')

    def recvUntil(self, conn, until):
        retbytearr = b''
        index = 0 - len(until)
        while True:
            rbytes = conn.recv(1)
            if rbytes == b'':
                break
            retbytearr += rbytes
            if retbytearr[index:] == until:
                break
        return retbytearr

    def recvFullHeader(self, conn):
        fullbytes = b''
        while True:
            rbytes = self.recvUntil(conn, b'\r\n')
            fullbytes += rbytes
            if rbytes == b'\r\n' or rbytes == b'':
                break
        return fullbytes

    def recvn(self, conn, n):
        toRecv = n
        hasRecv = 0
        bytearr = b''
        while True:
            if hasRecv >= toRecv:
                break
            bytearr += conn.recv(1)
            hasRecv += 1
        return bytearr

    def sendAll(self, conn, bytearr):
        toSend = len(bytearr)
        while toSend > 0:
            toSend -= conn.send(bytearr)
            bytearr = bytearr[0-toSend:] 


from sunyata.rpc.transport import TcpTransport, UdpTransport, HttpTransport, ClientUdpTransport, RpcTransport
from sunyata.rpc.serialize import BinarySerialize, JsonSerialize
from multiprocessing import cpu_count


class RpcProtocal(object):

    __slots__ = ('transport')

    def __init__(self):
        self.transport = RpcTransport()

    def serialize(self, obj):
        serializer = None
        if self.serializeType == 'bin':
            serializer = BinarySerialize()
        elif self.serializeType == 'json':
            serializer = JsonSerialize()
        if serializer == None:
            raise Exception('unknown serializeType')
        return serializer.serialize(obj)

    def unserialize(self, msg):
        serializer = None
        if self.serializeType == 'bin':
            serializer = BinarySerialize()
        elif self.serializeType == 'json':
            serializer = JsonSerialize()
        if serializer == None:
            raise Exception('unknown serializeType')
        return serializer.unserialize(msg) 

    def parseRequest(self, package):
        func = package['func']
        args = package['args']
        kwargs = package['kwargs']
        return func, args, kwargs


class UdpProtocal(RpcProtocal):

    def __init__(self):
        pass


class HttpProtocal(RpcProtocal):

    def __init__(self, host, port, worker = cpu_count(), serializeType = 'bin', timeout = 10, poolConnection=5, poolMaxSize = 20, maxRetries = 3):
        RpcProtocal.__init__(self)
        self.host = host
        self.port = port
        self.worker = worker
        self.serializeType = serializeType
        self.timeout = timeout
        self.transport = HttpTransport(host, port, timeout, poolConnection, poolMaxSize, maxRetries)


class TcpProtocal(RpcProtocal):
    
    def __init__(self, host, port, serializeType = 'bin', timeout = 10):
        RpcProtocal.__init__(self)
        self.serializeType = serializeType
        self.timeout = timeout
        self.transport = TcpTransport(host, port, timeout)


class UdpProtocal(RpcProtocal):

    def __init__(self, host, port, serializeType = 'bin', timeout = 10):
        RpcProtocal.__init__(self)
        self.serializeType = serializeType
        self.timeout = timeout
        self.transport = UdpTransport(host, port)


class ClientUdpProtocal(UdpProtocal):

    def __init__(self, host, port, serializeType = 'bin', timeout = 10):
        UdpProtocal.__init__(self, host, port, serializeType=serializeType, timeout=timeout)
        self.serializeType = serializeType
        self.host = host
        self.port = port
        self.timeout = timeout
        self.transport = ClientUdpTransport(host, port, timeout)
    
    def newTransport(self):
        self.transport.close()
        self.transport = ClientUdpTransport(self.host, self.port, self.timeout)


class Encrypt(object):

    PRIVATE_KEY = b'sunyata'

    @classmethod
    def encrypt(cls, bytearr):
        return cls.PRIVATE_KEY + bytearr

    @classmethod
    def unencrypt(cls, bytearr):
        keyLength = len(cls.PRIVATE_KEY)
        if len(bytearr) < keyLength:
            return False, b''
        header = bytearr[:keyLength]
        if header != cls.PRIVATE_KEY:
            return False, b''
        return True, bytearr[keyLength:]


#coding=utf-8
'''
Usage:

h = HashTable()
h.set("a", "1")
h.set("b", "2")
print(len(h))
print(h.get("a"))
h["abc"] = "abc"
print(h["abc"])
print(h.hasKey("a"))
for k, v in h:
    print(k, v)
for k, v in h.items():
    print(k, v)
'''


class HashTable(object):
    def __init__(self, size=99999):
        self.size = size
        self.hashList = [list() for _ in range(size)]

    def set(self, key, value):
        hashKey = hash(key) % self.size
        for item in self.hashList[hashKey]:
            if item[0] == key:
                item[1] = value
                return
        self.hashList[hashKey].append([key, value])

    def get(self, key):
        hashKey = hash(key) % self.size
        for item in self.hashList[hashKey]:
            if item[0] == key:
                return item[1]
        raise KeyError

    def hasKey(self, key):
        try:
            self.get(key)
        except KeyError:
            return False
        return True

    def __len__(self):
        length = 0
        for item in self.hashList:
            length = length + len(item)
        return length

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def __iter__(self):
        return self.items()

    def items(self):
        kvList = []
        for item in self.hashList:
            if len(item) > 0:
                kvList = kvList + item
        return iter(kvList)



from dataclasses import dataclass, field


@dataclass
class RequestClient(object):
    remoteAddr: str = field(default='')
    port: int = field(default=0)


class HttpRequest(object):

    def __init__(self):
        super().__init__()
        self.method = ''
        self.uri = ''
        self.httpVersion = ''
        self.body = ''
        self.headers = {}
        self.data = {}
        self.scheme = 'http'
        self.queryString = ''
        self.client = RequestClient()


    def setBody(self, body):
        self.body = body


import pickle
import ujson
from abc import ABCMeta, abstractmethod


class RpcSerialize(object):

    __metaclass__ = ABCMeta
    
    @abstractmethod
    def serialize(cls, obj) -> bytes:
        pass

    @abstractmethod
    def unserialize(cls, bytearr):
        pass


class BinarySerialize(RpcSerialize):

    @classmethod
    def serialize(cls, obj) -> bytes:
        return pickle.dumps(obj)

    @classmethod
    def unserialize(cls, bytearr):
        return pickle.loads(bytearr)


class JsonSerialize(RpcSerialize):

    @classmethod
    def serialize(cls, obj) -> bytes:
        return ujson.dumps(obj)

    @classmethod
    def unserialize(cls, bytearr):
        return ujson.loads(bytearr)


import asyncio

class EventLoop(object):
    @classmethod
    def runUntilComplete(self, asyncFunc):
        tasks = [ asyncFunc ]
        loop = asyncio.new_event_loop() 
        asyncio.set_event_loop(loop)
        loop = asyncio.get_event_loop()
        loop.run_until_complete( asyncio.gather(*tasks) )
        loop.run_forever()
        loop.close()


#coding=utf-8
'''
Usage:
bloom_filter = BloomFilter()
urls = ['www.baidu.com','mathpretty.com','sina.com']
urls_check = ['mathpretty.com','zhihu.com', 'sina.com', 'ali.com']
for url in urls: bloom_filter.add(url)
for url_check in urls_check:
    result = bloom_filter.contains(url_check)
    print('url : ',url_check,' contain : ',result)
'''

import crypt


class BloomFilter(object):
    def __init__(self, bitsize=30000):
        self.bitarr = [0 for _ in range(bitsize)]
        self.bitsize = bitsize

    def getPosList(self, string):
        posList = []
        for salt in ['41', '42', '43', '44', '45']:
            cryStr = crypt.crypt(string, salt)
            pos = hash(cryStr) % self.bitsize
            posList.append(pos)
        return posList

    def add(self, string):
        posList = self.getPosList(string)
        for pos in posList:
            self.bitarr[pos] = 1

    def contains(self, string):
        posList = self.getPosList(string)
        for pos in posList:
            if self.bitarr[pos] != 1:
                return False
        return True



import socket

def hostname():
    return socket.gethostname()

def local_ip():
    try:
        ip = socket.gethostbyname(socket.gethostname())
        if ip == '127.0.0.1':
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(('8.8.8.8', 80))
            ip = s.getsockname()[0]
    except:
        ip = '127.0.0.1'
    return ip

def bytes2str(bytes):
    return str(bytes, encoding = 'utf-8')

def str2bytes(string):
    return string.encode(encoding="utf-8")


'''
example:

from sunyata.redislock import get_redis_lock
import time

rlock = get_redis_lock('192.168.1.1', 6379, 'abcd1234')

while 1:
    if rlock.acquireLock():
        print('accquire lock succed')
        time.sleep(25)
        print('oper finish')
        rlock.releaseLock()
        print('release lock')
        time.sleep(2)
    else:
        time.sleep(2)
        print('wait lock')
'''

import redis
import uuid

class RedisLock(object):
    
    def __init__(self, redisConn, expire=30):
        super().__init__()
        self.redisConn = redisConn
        self.key = 'redis_lock_key'
        self.expire = expire
        self.lastvalue = str(uuid.uuid1())

    def acquireLock(self):
        ret = self.redisConn.setnx(self.key, self.lastvalue)
        if ret:
            self.redisConn.expire(self.key, self.expire)
            return True
        return False

    def releaseLock(self):
        value = self.redisConn.get(self.key)
        if value == None:
            return False
        if value != self.lastvalue:
            return False
        self.redisConn.delete(self.key)
        return True

def get_redis_lock(host,port,password):
    pool = redis.ConnectionPool(host=host, port=port, decode_responses=True, password=password)
    rconn = redis.Redis(connection_pool=pool)
    rlock = RedisLock(rconn)
    return rlock


#coding=utf-8

'''
Usage:

from sunyata.algorithm.trie import TrieNode, Trie
import random
arr = [u'运维', u'运维开发', u'运营']
print('arr', arr)
t = Trie()
for elem in arr:
    t.append(elem)
words = t.getWords()
print(words)
isExists = t.isContains('abcv')
print(isExists)
t.update(u'运维开发', u'运气')
words = t.getWordsByKw('运')
print(words)

'''

class TrieNode(object):
    
    def __init__(self):
        self.char = ''
        self.pre = None
        self.isRoot = False
        self.childMap = {}
        self.data = None
        self.isWordEnd = False
    
    def isLeaf(self):
        if len(self.childMap) <= 0:
            return True
        return False
    
    def isRoot(self):
        return self.isRoot
    
    def childNodes(self):
        nodes = []
        for _, node in (self.childMap).items():
            nodes.append(node)
        return nodes
    
    def hasChild(self):
        childNodes = self.childNodes()
        if len(childNodes) <= 0:
            return False
        return True


class Trie(object):
    
    def __init__(self):
        self.root = TrieNode()
        self.root.isRoot = True
        self.curString = ''
        self.words = []
        
    def getRootNode(self):
        return self.root
    
    def isHasChildNodeByChar(self, node, char):
        if char in node.childMap:
            return True
        return False
        
    def getChildNodeByChar(self, node, char):
        return node.childMap.get(char)
    
    def append(self, string):
        curnode = self.root
        linkstr = ''
        for char in string:
            linkstr = linkstr + char
            if self.isHasChildNodeByChar(curnode, char):
                curnode = self.getChildNodeByChar(curnode, char)
            else:
                childNode = TrieNode()
                childNode.char = char
                childNode.pre = curnode
                childNode.isRoot = False
                childNode.data = linkstr
                curnode.childMap[char] = childNode
                curnode = childNode
        curnode.isWordEnd = True
                
    def getNodeWord(self, node):
        words = []
        if node.isWordEnd:
            words.append(node.data)
        if node.hasChild():
            for childNode in node.childNodes():
                words = words + self.getNodeWord(childNode)
        return words
                
    def getWords(self):
        words = self.getNodeWord(self.root)
        return words
    
    def isContains(self, kw):
        curnode = self.root
        for char in kw:
            if self.isHasChildNodeByChar(curnode, char):
                curnode = self.getChildNodeByChar(curnode, char)
            else:
                return False
        return True
    
    def getWordsByKw(self, kw):
        words = []
        curnode = self.root
        for char in kw:
            if self.isHasChildNodeByChar(curnode, char):
                curnode = self.getChildNodeByChar(curnode, char)
            else:
                return words
        words = self.getNodeWord(curnode)
        return words
    
    def remove(self, string):
        curnode = self.root
        for char in string:
            if self.isHasChildNodeByChar(curnode, char):
                curnode = self.getChildNodeByChar(curnode, char)
            else:
                return
        if curnode.hasChild():
            curnode.isWordEnd = False
            return
        while curnode != self.root:
            if not curnode.hasChild():
                if not curnode.pre.isWordEnd:
                    del curnode.pre.childMap[curnode.char]
                    curnode = curnode.pre
                else:
                    break
            else:
                break
            
    def update(self, string, newstring):
        self.remove(string)
        self.append(newstring)


from sunyata.util import str2bytes, bytes2str
import ujson

class HttpResponse(object):

    __slots__ = ('httpVersion', 'status', 'headers', 'body')

    def __init__(self, status, body) -> None:
        super().__init__()
        self.httpVersion = 'HTTP/1.1'
        self.status = status
        self.headers = {}
        self.body = body

    def genHeaders(self):
        self.headers['Content-Type'] = 'text/html'
        self.headers['Content-Length'] = len(self.body)

    def toBytes(self):
        resLine = f"{self.httpVersion} {self.status.code} {self.status.msg}"
        self.genHeaders()
        resHeaders = ''
        for k, v in self.headers.items():
            resHeaders += f"{k}: {v}\r\n"
        resHeaders += "\r\n"
        resBody = self.body
        if type(resBody) == str:
            resBody = str2bytes(resBody)
        res = str2bytes(resLine + resHeaders) + resBody
        return res
    
    def responseBody(self):
        resBody = self.body
        if type(resBody) == str:
            resBody = str2bytes(resBody)
        return resBody
    
    @classmethod
    def responseConvertSetToJson(cls, resp):
        return ujson.encode(list(resp))
    
    @classmethod
    def responseConverBytesToStr(cls, resp):
        return bytes2str(resp)
    
    @classmethod
    def responseConvertToJson(cls, resp):
        return ujson.encode(resp)
    
    @classmethod
    def responseConvertToBytes(cls, resp):
        return bytes(resp)
    
    @classmethod
    def responseConvertToStr(cls, resp):
        return str(resp)
    
    @classmethod
    def responseConvertRaw(self, resp):
        return resp


'''
from sunyata.orm import Model, IntField, CharField

#orm init
Model.init('127.0.0.1', 3306, 'root', '', 'test2', min_conn_num=10)


#define a model
class Nation(Model):
    tableName = 'nation' #required
    primaryKey = 'id'    #required
    id = IntField()      #field type int
    name = CharField()   #field type char


#create
nation = Nation(id = 30, name = 'dqdsaw')
nation.create()

#read
obj = Nation.filter('id', '=', 30).get()
print(obj, obj.id, obj.name)
objs = Nation.all()

#update
obj.name = 'hahaha'
obj.update()
obj = Nation.filter('id', '=', 30).get()
print(obj, obj.id, obj.name)

#delete
obj.delete()
obj = Nation.filter('id', '=', 30).first()
print(obj)
'''

from sunyata.db import PoolDB
from sunyata.db import Orm

class OrmObject(object):

    def __init__(self):
        pass


class OrmModel(OrmObject):
    _connection = None
    tableName = ''
    primaryKey = 'id'

    def __init__(self):
        self.fieldMap = {}

    @classmethod
    def init(cls, host, port, user, passwd, dbName, connect_timeout=10, read_timeout=10, min_conn_num=10,):
        cls._connection = PoolDB(
            host, port, user, passwd, dbName, 
            connect_timeout=connect_timeout, read_timeout=read_timeout, 
            min_conn_num=min_conn_num)
        cls._connection.connect()

    @classmethod
    def getOrm(cls):
        return Orm(cls._connection).table(cls.tableName)

    @classmethod
    def getConnection(cls):
        return cls.getOrm()

    @classmethod
    def getRawConnection(cls):
        return cls._connection


class OrmTypeObject(OrmObject):

    def __init__(self, comment = ''):
        pass

class IntField(OrmTypeObject):
    typeName = 'int'

class CharField(OrmTypeObject):
    typeName = 'char'

class FloatField(OrmTypeObject):
    typeName = 'float'


class Model(OrmModel):

    def __init__(self, **args):
        super().__init__()
        for k, v in args.items():
            if not hasattr(self, k):
                raise Exception('no attr' + str(k))
            self.fieldMap[k] = v

    def create(self):
        return self.getOrm().data(self.fieldMap).insert()

    @classmethod
    def filter(cls, field, cmp, val, raw = False):
        rows = cls.getOrm().where(field, cmp, val).get()
        if raw:
            return rows
        querySet = OrmQuerySet()
        for row in rows:
            obj = cls()
            for k, v in row.items():
                if hasattr(obj, k):
                    setattr(obj, k, v)
                obj.fieldMap[k] = v
            querySet.append(obj)
        return querySet

    def update(self):
        data = {}
        for k in self.fieldMap.keys():
            data[k] = getattr(self, k)
        return self.getOrm().data(data).where(self.primaryKey, '=', getattr(self, self.primaryKey)).update()

    def delete(self):
        return self.getOrm().where(self.primaryKey, '=', getattr(self, self.primaryKey)).delete()

    @classmethod
    def all(cls, raw = False):
        rows = cls.getOrm().get()
        if raw:
            return rows
        querySet = OrmQuerySet()
        for row in rows:
            obj = cls()
            for k, v in row.items():
                if hasattr(obj, k):
                    setattr(obj, k, v)
                obj.fieldMap[k] = v
            querySet.append(obj)
        return querySet


class OrmQuerySet(Model):

    __slots__ = ['set', 'index']
    
    def __init__(self):
        self.set = []
        self.index = 0

    def append(self, elem):
        self.set.append(elem)
        
    def __iter__(self):
        return self

    def __next__(self):
        try:
            elem = self.set[ self.index ]
            self.index = self.index + 1
            return elem
        except:
            raise StopIteration('')

    def objects(self):
        return self.set

    def delete(self):
        for elem in self.set:
            OrmQuerySet.tableName = elem.tableName
            self.getOrm().where(elem.primaryKey, '=', getattr(elem, elem.primaryKey)).delete()

    def first(self):
        if len(self.set) <= 0:
            return None
        return self.set[0]

    def get(self):
        if len(self.set) <= 0:
            raise Exception('no result')
        return self.set[0]

    def count(self):
        return len(self.set)


class HttpRouter(object):

    __slots__ = ('path', 'methods', 'func')

    def __init__(self) -> None:
        super().__init__()
        self.path = ''
        self.methods = None
        self.func = None

    def getFunc(self):
        return self.func


'''
usage:

header = ['name', 'age']
rows = [ ['wangermazi', 25], ['zhangsan', 11], ['lisi', 30] ]
tbWriter = TableWriter(header, rows)
tbWriter.dump()
string = tbWriter.getString()
print(string)

+--------------+-------+
|  name        |  age  |
+--------------+-------+
|  wangermazi  |  25   |
|  zhangsan    |  11   |
|  lisi        |  30   |
+--------------+-------+
'''


class TableWriter:

    __slots__ = ['_maxCol', '_header', '_rows', '_colMaxLengthList', '_lWidth', '_rWidth']

    def __init__(self, header=[], rows=[], lWidth=2, rWidth=2):
        if type(header) != list or type(rows) != list:
            raise Exception('param headers, rows must be list type')

        self._maxCol = len(header)
        for row in rows:
            if len(row) > self._maxCol:
                self._maxCol = len(row)

        self._header = header
        self._rows = rows
        self._colMaxLengthList = self._getColsMaxLength()
        self._lWidth = lWidth
        self._rWidth = rWidth

    def _getColsMaxLength(self):
        allRows = []
        allRows.append(self._header)
        for row in self._rows:
            allRows.append(row)

        colMaxLengthList = []
        for i in range(self._maxCol):
            colMaxLength = 0

            for row in allRows:
                try:
                    fieldLength = len(str(row[i]))
                except:
                    continue
                if fieldLength > colMaxLength:
                    colMaxLength = fieldLength

            colMaxLengthList.append(colMaxLength)

        return colMaxLengthList

    def _genFieldStr(self, string, maxLength):
        string = str(string)
        restSpaceLength = maxLength - len(string)
        ret = ''.join([' ' for i in range(self._lWidth)])
        ret = ret + string + ''.join(
            [' ' for i in range(restSpaceLength + self._rWidth)])
        return ret

    def getString(self):
        headerFieldList = []
        for i in range(len(self._header)):
            fieldStr = self._genFieldStr(self._header[i],
                                         self._colMaxLengthList[i])
            headerFieldList.append(fieldStr)
        headerStr = '|' + '|'.join(headerFieldList) + '|' + "\n"

        rowListStr = ''
        for row in self._rows:
            rowFieldList = []
            for i in range(len(row)):
                fieldStr = self._genFieldStr(row[i], self._colMaxLengthList[i])
                rowFieldList.append(fieldStr)
            rowStr = '|' + '|'.join(rowFieldList) + '|'
            rowListStr = rowListStr + rowStr + "\n"

        indentList = []
        for maxLength in self._colMaxLengthList:
            fieldStr = ''.join(
                ['-' for i in range(maxLength + self._lWidth + self._rWidth)])
            indentList.append(fieldStr)
        indentStr = '+' + '+'.join(indentList) + '+' + "\n"

        ret = ''
        if len(self._header) > 0:
            ret = ret + indentStr + headerStr + indentStr
        if len(self._rows) > 0:
            ret = ret + rowListStr + indentStr

        return ret

    def dump(self):
        print(self.getString())


import sys
try:
    from imp import reload
except:
    pass
reload(sys)
try:
    sys.setdefaultencoding('utf-8')
except:
    pass

import logging
from logging.handlers import WatchedFileHandler
from logging import handlers


class Log:

    __slots__ = [
        '__path', '__switch', '__fileHandler', '__logger', '_errorCallBack',
        '_warningCallBack', '_isOutput'
    ]

    __fmt = '%(asctime)s [%(levelname)s] [p:%(process)d] [t:%(thread)d] [%(threadName)s] %(message)s'
    __shortFmt = '%(asctime)s-%(levelname)s-%(message)s'
    __level = logging.DEBUG

    def __init__(self, path, isRotate=True, logSaveDays=10, isShortLog=False):
        self.__path = path
        self.__switch = True
        fmter = None
        if isShortLog:
            fmter = logging.Formatter(Log.__shortFmt)
        else:
            fmter = logging.Formatter(Log.__fmt)
        if isRotate:
            self.__fileHandler = handlers.TimedRotatingFileHandler(
                self.__path,
                when='D',
                interval=1,
                backupCount=logSaveDays,
                encoding='utf-8')
        else:
            self.__fileHandler = WatchedFileHandler(self.__path)
        self.__fileHandler.setFormatter(fmter)
        self.__logger = logging.getLogger(path)
        self.__logger.addHandler(self.__fileHandler)
        self.__logger.setLevel(Log.__level)
        self._errorCallBack = None
        self._warningCallBack = None
        self._isOutput = False

    def setOutput(self, tag):
        self._isOutput = tag

    def getOutput(self):
        return self._isOutput

    def setSwitch(self, switch):
        self.__switch = switch

    def getSwitch(self):
        return self.__switch

    def getPath(self):
        return self.__path

    def debug(self, *args):
        logInfo = ' '.join(str(item) for item in args)
        if self.__switch: self.__logger.debug(loginfo)
        if self._isOutput: print(logInfo)

    def info(self, *args):
        logInfo = ' '.join(str(item) for item in args)
        if self.__switch: self.__logger.info(logInfo)
        if self._isOutput: print(logInfo)

    def warning(self, *args):
        logInfo = ' '.join(str(item) for item in args)
        if self.__switch: self.__logger.warning(logInfo)
        if self._isOutput: print(logInfo)
        try:
            if self._warningCallBack: self._warningCallBack(logInfo)
        except:
            pass

    def error(self, *args):
        logInfo = ' '.join(str(item) for item in args)
        if self.__switch: self.__logger.error('\033[31m' + logInfo + "\033[0m")
        if self._isOutput: print('\033[31m' + logInfo + '\033[0m')
        try:
            if self._errorCallBack: self._errorCallBack(logInfo)
        except:
            pass

    def fatal(self, *args):
        logInfo = ' '.join(str(item) for item in args)
        print(logInfo)
        import os
        os._exit(1)

    def setErrorCallBack(self, func):
        self._errorCallBack = func

    def setWarningCallBack(self, func):
        self._warningCallBack = func


import lz4.block as lb

class Compress(object):

    level = 3

    DEBUG = False

    @classmethod
    def compress(cls, bytearr: bytes):
        compressed = lb.compress(bytearr)
        if cls.DEBUG:
            print('debug, do compress', 'orig len', len(bytearr), 'compress len', len(compressed))
        return compressed

    @classmethod
    def decompress(cls, bytearr):
        if cls.DEBUG:
            print('debug, do decompress')
        origin = lb.decompress(bytearr)
        return origin


import functools
import redis
import pickle

FUNC_CACHE_MAP = {}

def _get_key(funcname, *args, **kwargs):
    key = 'sunyata_cache_func:'+funcname + ':'
    if len(args) > 0:
        key = key + str(list(args)) + ':'
    if len(kwargs) > 0:
        key = key + str(dict(kwargs))
    return key

def cache_func(*args, **kwargs):
    def function(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            global FUNC_CACHE_MAP
            key = _get_key(func.__name__, *args, **kwargs)
            if key in FUNC_CACHE_MAP:
                return FUNC_CACHE_MAP[key]
            ret = func(*args, **kwargs)
            FUNC_CACHE_MAP[key] = ret
            return ret
        return wrapper
    return function

def get_redis_connection(connstr):
    host, port, db = connstr.split(':')
    conn = redis.Redis(host=host,port=port,db=db)
    return conn

def redis_cache_func(connstr='127.0.0.1:6379:1', expireSeconds=3600*24):
    conn = get_redis_connection(connstr)
    def function(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            key = _get_key(func.__name__, *args, **kwargs)
            val = conn.get(key)
            if val != None:
                return pickle.loads(val)
            ret = func(*args, **kwargs)
            val = pickle.dumps(ret)
            conn.set(key, val)
            conn.expire(key, expireSeconds)
            return ret
        return wrapper
    return function


class HttpStatus(object):
    code = 0
    msg = ''


class HttpStatus200(HttpStatus):
    code = 200
    msg = 'OK'


class HttpStatus400(HttpStatus):
    code = 400
    msg = 'Bad Request'


class HttpStatus401(HttpStatus):
    code = 401
    msg = 'Unauthorized'
        

class HttpStatus403(HttpStatus):
    code = 403
    msg = 'Forbidden'


class HttpStatus404(HttpStatus):
    code = 404
    msg = 'Not Found'


class HttpStatus405(HttpStatus):
    code = 405
    msg = 'Method Not Allowed'


class HttpStatus500(HttpStatus):
    code = 500
    msg = 'Internal Server Error'


class HttpStatus503(HttpStatus):
    code = 503
    msg = 'Server Unavailable'


from sunyata.consul import ConsulApi, Instance
import time
import threading
import asyncio
from sunyata.etcd import EtcdApi


class DiscoveryConfig(object):

    def __init__(self, serviceName, consulHost = None, consulPort = None, etcdHost=None, etcdPort=None, serviceHost = '', servicePort = 0, consulToken=''):
        self.consulHost = consulHost
        self.consulPort = consulPort
        self.consulToken = consulToken
        self.serviceName = serviceName
        self.serviceHost = serviceHost
        self.servicePort = servicePort
        self.etcdHost = etcdHost
        self.etcdPort = etcdPort


class RpcDiscovery(object):

    def __init__(self, consulHost=None, consulPort=None, consulToken='', etcdHost=None, etcdPort=None):
        self.consulHost = consulHost
        self.consulPort = consulPort
        self.consulToken = consulToken
        self.etcdHost = etcdHost
        self.etcdPort = etcdPort
        self.api = None
        if self.consulHost and self.consulPort:
            self.api = ConsulApi(consulHost, consulPort, consulToken)
        if self.etcdHost and self.etcdPort:
            self.api = EtcdApi(etcdHost, etcdPort)
        self.instanceList = []
        self.heartbeatInterval = 15

    def doHeartbeat(self, service, address, port, interval):
        while 1:
            try:
                self.api.ttlHeartbeat(service, address, port)
            except:
                pass
            time.sleep(interval)
    
    async def asyncDoHeartbeat(self, service, address, port, interval):
        while 1:
            try:
                self.api.ttlHeartbeat(service, address, port)
            except:
                pass
            await asyncio.sleep(interval)

    def regist(self, service, address, port, ttlHeartBeat = True):
        self.api.registService(serviceName = service, address = address, port=port)
        if ttlHeartBeat:
            t = threading.Thread(target=self.doHeartbeat, args=(service, address, port, self.heartbeatInterval))
            t.setDaemon(True)
            t.start()

    async def asyncRegist(self, service, address, port, ttlHeartBeat = True):
        self.api.registService(serviceName = service, address = address, port=port)
        await self.asyncDoHeartbeat(service, address, port, self.heartbeatInterval)

    def getInstanceList(self, service):
        instanceList = self.api.getServiceInstanceList(service)
        return instanceList


from sunyata.http.request import HttpRequest
from sunyata.http.response import HttpResponse
from sunyata.http.router import HttpRouter
from sunyata.util import bytes2str
import ujson

class HttpFactory(object):

    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def genLineHeaderBody(self, lines):
        getHeader = 1
        reqHeaders = []
        reqBody = b''
        reqLine = lines[0]
        for l in lines[1:]:
            if not l:
                getHeader = 0
            if getHeader:
                reqHeaders.append(l)
            else:
                reqBody = reqBody + l
        return reqLine, reqHeaders, reqBody

    @classmethod
    def parseRequestData(cls, uri, body):
        requestData = {}
        if uri:
            try:
                kvPairs = uri.split(b'?')[1].split(b'&')
                for kvPair in kvPairs:
                    k, v = kvPair.split(b'=')
                    requestData[bytes2str(k)] = bytes2str(v)
            except:
                pass
        if body:
            try:
                bodydic = ujson.decode(body)
                requestData.update(bodydic)
            except:
                pass
        return requestData
        
    @classmethod
    async def genHttpRequest(cls, requestStream):
        linesStr = await requestStream.reader.read(requestStream.bufsize)
        req = HttpRequest()
        lines = linesStr.split(b"\r\n")
        reqLine, reqHeaders, reqBody = cls.genLineHeaderBody(lines)
        method, uri, httpVersion = reqLine.split(b' ')
        req.method = bytes2str(method)
        req.httpVersion = bytes2str(httpVersion)
        req.uri = bytes2str( uri.split(b'?')[0] )
        req.body = reqBody
        for header in reqHeaders:
            if header == b'':
                continue
            k, v = header.split(b': ')
            try:
                req.headers[bytes2str(k)] = bytes2str(v)
            except:
                req.headers[str(k)] = str(v)
        req.data = cls.parseRequestData(uri, req.body)
        contentLength = int(req.headers.get('Content-Length', 0))
        hasRead = len(req.body)
        toRead = contentLength - hasRead
        while toRead:
            rdata = requestStream.reader.read(toRead)
            if not rdata:
                raise Exception('peer closed')
            req.body = req.body + rdata
            hasRead = hasRead + len(rdata)
            toRead = contentLength - hasRead
        return req

    @classmethod
    def genHttpResponse(cls, status, body = ''):
        return HttpResponse(status, body)

    @classmethod
    def genHttpRouter(cls, path, func, methods):
        httpRouter = HttpRouter()
        httpRouter.path = path
        httpRouter.func = func
        httpRouter.methods = methods
        return httpRouter


from sunyata.rpc.protocal import TcpProtocal, UdpProtocal, HttpProtocal, RpcProtocal, ClientUdpProtocal
from sunyata.rpc.exception import FuncNotFoundException
from sunyata.rpc.discovery import DiscoveryConfig
from sunyata.rpc.discovery import RpcDiscovery
from sunyata.wrap import retryTimes
import socket

class Address(object):

    __slots__ = ('host', 'port')

    def __init__(self, host, port):
        self.host = host
        self.port = port


class Callable(object):

    def __init__(self, rpcClient, remoteMethodName):
        self.rpcClient = rpcClient
        self.remoteMethodName = remoteMethodName

    def __call__(self, *args, **kwargs):
        return self.rpcClient.call(self.remoteMethodName, *args, **kwargs)

    def __getattr__(self, key):
        if key in dir(self):
            return getattr(self, key)
        return Callable(self.rpcClient, self.remoteMethodName+'.'+f"{key}")


class RpcClient(object):

    __slots__ = ('discoveryConfig', 'discovery', 'instanceIndex', 'protocalMap', 'isSync', 'syncInterval', 'protocal', 'servers', 'serversCount', 'serversProtocalMap', 'serverIndex', 'lastServer', 'timeout') 

    def __init__(self):
        self.discoveryConfig = None
        self.discovery = None
        self.instanceIndex = 0
        self.protocalMap = {}
        self.isSync = False
        self.syncInterval = 30
        self.protocal = RpcProtocal()
        self.servers = []
        self.serversCount = len(self.servers)
        self.serversProtocalMap = {}
        self.serverIndex = 0
        self.lastServer = ''
        self.timeout = 10

    def getInstance(self):
        instanceList = self.discovery.getInstanceList(self.discoveryConfig.serviceName)
        instanceLength  = len(instanceList)
        if instanceLength == 0:
            raise Exception('no instance found')
        index = self.instanceIndex % instanceLength
        self.instanceIndex = self.instanceIndex + 1
        return instanceList[index]

    def refreshProtocal(self):
        if self.discovery:
            instance = self.getInstance()
            self.protocal = self.getProtocalInstance(instance)
        if self.servers:
            self.serversCount = len(self.servers)
            index = self.serverIndex % self.serversCount
            self.serverIndex = self.serverIndex + 1
            server = self.servers[ index ]
            self.lastServer = server
            self.protocal = self.serversProtocalMap[ server ]

    def getProtocalInstance(self, instance):
        #overide
        pass

    def setDiscoveryConfig(self, config: DiscoveryConfig):
        self.discoveryConfig = config
        self.discovery = RpcDiscovery(self.discoveryConfig.consulHost, self.discoveryConfig.consulPort, self.discoveryConfig.consulToken, self.discoveryConfig.etcdHost, self.discoveryConfig.etcdPort)

    def close(self):
        self.protocal.transport.close()

    def __del__(self):
        self.close()

    def beforeCall(self):
        self.refreshProtocal()

    def debugGetLastServer(self):
        return self.lastServer

    @retryTimes(retryTimes=3)
    def call(self, func, *args, **kwargs):
        pass

    def __getattr__(self, key):
        if key in dir(self):
            return getattr(self, key)
        return Callable(self, f"{key}")


class TcpRpcClient(RpcClient):

    def __init__(self, host = '', port = 0, servers=[], timeout=10, keepconnect = True):
        RpcClient.__init__(self)
        self.host = host
        self.port = port
        self.keepconnect = keepconnect
        self.protocal = TcpProtocal(host, port, timeout=timeout)
        self.servers = servers
        self.timeout = timeout
        if len(self.servers) == 0 and self.host != '' and self.port != 0:
            self.servers = [ host+':'+str(port) ]
        for server in self.servers:
            host, port = server.split(':')
            self.serversProtocalMap[ server ] = TcpProtocal(host, int(port), timeout = timeout)

    @retryTimes(retryTimes=3)
    def call(self, func, *args, **kwargs):
        try:
            self.beforeCall()
            self.protocal.transport.connect()
            package = {
                'func' : func,
                'args' : args,
                'kwargs' : kwargs,
            }
            msg = self.protocal.serialize(package)
            self.protocal.transport.send(msg)
            respmsg = self.protocal.transport.recv()
            resp = self.protocal.unserialize(respmsg)
            if isinstance(resp, Exception):
                if not self.keepconnect: self.protocal.transport.close()
                raise resp
            if not self.keepconnect: self.protocal.transport.close()
            return resp
        except socket.timeout as ex:
            self.protocal.transport.close()
            raise ex

    def getProtocalInstance(self, instance):
        key = "%s:%s:%s" % (instance.service, instance.address, instance.port)
        if key in self.protocalMap:
            return self.protocalMap.get(key)
        self.protocalMap[key] =  TcpProtocal(instance.address, instance.port)
        return self.protocalMap[key]


class UdpRpcClient(RpcClient):

    def __init__(self, host = '', port = 0, servers=[], timeout = 10):
        RpcClient.__init__(self)
        self.host = host
        self.port = port
        self.protocal = ClientUdpProtocal(host, port, timeout=timeout)
        self.servers = servers
        self.timeout = timeout
        if len(self.servers) == 0 and self.host != '' and self.port != 0:
            self.servers = [ host+':'+str(port) ]
        for server in self.servers:
            host, port = server.split(':')
            self.serversProtocalMap[ server ] = ClientUdpProtocal(host, int(port), timeout=timeout)

    @retryTimes(retryTimes=3)
    def call(self, func, *args, **kwargs):
        try:
            self.beforeCall()
            package = {
                'func' : func,
                'args' : args,
                'kwargs' : kwargs
            }
            msg = self.protocal.serialize(package)
            self.protocal.transport.send(msg)
            respmsg, _ = self.protocal.transport.recv()
            resp = self.protocal.unserialize(respmsg)
            if isinstance(resp, Exception):
                raise resp
            return resp
        except socket.timeout as ex:
            self.protocal.newTransport()
            raise ex

    def getProtocalInstance(self, instance):
        key = "%s:%s:%s" % (instance.service, instance.address, instance.port)
        if key in self.protocalMap:
            return self.protocalMap.get(key)
        self.protocalMap[key] =  UdpProtocal(instance.address, instance.port)
        return self.protocalMap[key]


class HttpRpcClient(RpcClient):

    def __init__(self, host = '', port = 0, servers = [], timeout = 10):
        RpcClient.__init__(self)
        self.host = host
        self.port = port
        self.protocal = HttpProtocal(host, port, timeout=timeout)
        self.servers = servers
        self.timeout = timeout
        if len(self.servers) == 0 and self.host != '' and self.port != 0:
            self.servers = [ host+':'+str(port) ]
        for server in self.servers:
            host, port = server.split(':')
            self.serversProtocalMap[ server ] = HttpProtocal(host, int(port), timeout=timeout)

    @retryTimes(retryTimes=3)
    def call(self, func, *args, **kwargs):
        self.beforeCall()
        package = {
            'func' : func,
            'args' : args,
            'kwargs' : kwargs,
        }
        msg = self.protocal.serialize(package)
        respmsg = self.protocal.transport.send(msg)
        resp = self.protocal.unserialize(respmsg)
        if isinstance(resp, FuncNotFoundException):
            raise resp
        return resp

    def getProtocalInstance(self, instance):
        key = "%s:%s:%s" % (instance.service, instance.address, instance.port)
        if key in self.protocalMap:
            return self.protocalMap.get(key)
        self.protocalMap[key] =  HttpProtocal(instance.address, instance.port)
        return self.protocalMap[key]


#coding=utf-8
'''
Usage:


lru = LRUCache(size = 10)
for i in range(10):
    key = str(i)
    lru.put(key, key) 
lru.dump()
for i in range(5):
    lru.get(str(i))
lru.dump()
for k, v in lru.items():
    print(k, v)
'''

import collections


class Node(object):
    def __init__(self, key, value):
        self.key = key
        self.value = value
        self.next = None


class LinkNode(object):
    def __init__(self):
        self.head = None
        self.tail = None
        self.length = 0

    def append(self, node):
        if self.head == None and self.tail == None:
            node.next = None
            self.head = node
            self.tail = node
        else:
            self.tail.next = node
            node.next = None
            self.tail = node
        self.length = self.length + 1

    def dump(self):
        node = self.head
        while node != None:
            print(node.key, node.value)
            node = node.next
        print('---------------------')

    def items(self):
        kvList = []
        node = self.head
        while node != None:
            kvList.append([node.key, node.value])
            node = node.next
        return iter(kvList)

    def find(self, key):
        node = self.head
        while node != None:
            if node.key == key:
                return node
            node = node.next
        return None

    def appendHead(self, node):
        node.next = self.head
        if self.head == None and self.tail == None:
            self.tail = node
        self.head = node
        self.length = self.length + 1

    def removeTail(self):
        if self.tail == None: return
        preNode = self.head
        curNode = self.head
        while 1:
            if curNode != self.tail:
                preNode = curNode
                curNode = curNode.next
            else:
                break
        preNode.next = None
        self.tail = preNode
        self.length = self.length - 1

    def removeNode(self, node):
        if self.head == None and self.tail == None:
            return
        if self.head == node:
            self.head = node.next
            self.length = self.length - 1
            return
        if self.tail == node:
            self.removeTail()
            return
        preNode = curNode = self.head
        while curNode != self.tail:
            if curNode.key == node.key:
                preNode.next = node.next
                self.length = self.length - 1
                return
            preNode = curNode
            curNode = curNode.next


class LRUCache(object):
    def __init__(self, size=9999):
        self.size = size
        self.link = LinkNode()

    def get(self, key):
        node = self.link.find(key)
        if node != None:
            self.link.removeNode(node)
            self.link.appendHead(node)
        else:
            raise KeyError

    def put(self, key, value):
        node = self.link.find(key)
        if node != None:
            self.link.removeNode(node)
            self.link.appendHead(node)
        else:
            node = Node(key, value)
            if (self.link.length + 1) <= self.size:
                self.link.appendHead(node)
            else:
                self.link.appendHead(node)
                self.link.removeTail()
        return self

    def length(self):
        return self.link.length

    def dump(self):
        return self.link.dump()

    def items(self):
        return self.link.items()



from sunyata.http.factory import HttpFactory
from sunyata.http.status import *
from sunyata.http.request_stream import RequestStream
from sunyata.http.response import HttpResponse
from traceback import format_exc
from types import MethodType,FunctionType
import asyncio
import uvloop
import inspect
asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


class RawHttpServer(object):

    routerMap = {}
    responseTypeConvertMap = {
        set: HttpResponse.responseConvertSetToJson,
        dict: HttpResponse.responseConvertToJson,
        list: HttpResponse.responseConvertToJson,
        tuple: HttpResponse.responseConvertToJson,
        bytes: HttpResponse.responseConvertRaw,
    }

    def __init__(self, bind = '0.0.0.0', port=9989):
        super().__init__()
        self.bind = bind
        self.port = port
    
    @classmethod
    def addRoute(cls, path, func, methods = None):
        if isinstance(func, FunctionType) or isinstance(func, MethodType):
            cls.routerMap[path] = HttpFactory.genHttpRouter(path, func, methods)
        elif isinstance(func, object):
            controllerObj = func()
            if not path.endswith('/'): path += '/'
            for classFunc in dir(controllerObj):
                if classFunc.startswith('_'):
                    continue
                cls.routerMap[path + classFunc] = HttpFactory.genHttpRouter(path + classFunc, getattr(controllerObj, classFunc), methods)

    async def handleRequest(self, httpRequest):
        router = self.routerMap.get(httpRequest.uri, None)
        if not router:
            return HttpFactory.genHttpResponse(HttpStatus404)
        if router.methods and httpRequest.method not in router.methods:
            return HttpFactory.genHttpResponse(HttpStatus405)
        try:
            if inspect.iscoroutinefunction(router.getFunc()):
                resp = await router.getFunc()(httpRequest)
            else:
                resp = router.getFunc()(httpRequest)
            respType = type(resp)
            convertMethod = self.responseTypeConvertMap.get(respType, HttpResponse.responseConvertToStr)
            resp = convertMethod(resp)
            return HttpFactory.genHttpResponse(HttpStatus200, resp)
        except Exception as ex:
            return HttpFactory.genHttpResponse(HttpStatus500, format_exc())
    
    async def acceptStream(self, reader, writer):
        try:
            requestStream = RequestStream(reader, writer)
            httpRequest = await HttpFactory.genHttpRequest(requestStream)
            httpResponse = await self.handleRequest(httpRequest)
        except Exception as ex:
            httpResponse = HttpFactory.genHttpResponse(HttpStatus500, str(ex))
        writer.write(httpResponse.toBytes())
        await writer.drain()
        writer.close()
    
    async def listenAndServe(self):
        tasklist = []
        server = await asyncio.start_server(self.acceptStream, self.bind, self.port)
        tasklist.append(server.serve_forever())
        for task in tasklist:
            await task

    def serve(self):
        print('http running on http://%s:%s' % (self.bind, self.port) )
        asyncio.run(self.listenAndServe())

    @classmethod
    def route(cls, path, methods = None):
        def wrapper(func):
            cls.addRoute(path, func, methods)
            return func
        return wrapper

route = RawHttpServer.route


'''
mysql conneciton pool
'''

import pymysql
from DBUtils.PooledDB import PooledDB

import sys
if sys.version[0:1] == '3': unicode = str


class PoolDB(object):

    __slots__ = (
        'host',
        'port',
        'user',
        'passwd',
        'dbName',
        'log',
        'connectTimeout',
        'readTimeout',
        'minConnNum',
    )

    def __init__(self,
                 host,
                 port,
                 user,
                 passwd,
                 dbName,
                 log=None,
                 connect_timeout=10,
                 read_timeout=None,
                 min_conn_num=10):
        self.host = host
        self.port = port
        self.user = user
        self.passwd = passwd
        self.dbName = dbName
        self.log = log
        self.connectTimeout = connect_timeout
        self.readTimeout = read_timeout
        self.minConnNum = min_conn_num
        self.charset = 'utf8'
        self.pool = None

    def connect(self):
        if self.pool == None:
            self.pool = PooledDB(pymysql,
                                 self.minConnNum,
                                 host=self.host,
                                 port=self.port,
                                 user=self.user,
                                 passwd=self.passwd,
                                 db=self.dbName,
                                 charset=self.charset,
                                 connect_timeout=self.connectTimeout,
                                 read_timeout=self.readTimeout)
            print('PoolDB init finish')

    def close(self):
        if self.pool != None:
            self.pool.close()
            self.pool = None

    def query(self, sql):
        conn = self.pool.connection()
        cur = conn.cursor(pymysql.cursors.DictCursor)
        cur.execute(sql)
        res = cur.fetchall()
        cur.close()
        conn.close()
        return res

    def update(self, sql):
        conn = self.pool.connection()
        cur = conn.cursor()
        effect_rows = cur.execute(sql)
        conn.commit()
        lastrowId = cur.lastrowid
        cur.close()
        conn.close()
        return effect_rows, lastrowId


class Orm:

    ORDER_ASC = 'asc'
    ORDER_DESC = 'desc'

    def __init__(self, sunyataDb):
        self.db = sunyataDb
        self.tableName = ''
        self.whereCon = ''
        self.updateData = None
        self.fields = []
        self.raw = ''

    def table(self, tableName):
        self.tableName = tableName
        return self

    def field(self, field_name):
        tmpFields = list(set( [ item.strip() for item in field_name.split(',') ] ))
        for tf in tmpFields:
            self.fields.append(tf)
        self.fields = list(set(self.fields))
        return self

    def sql(self, raw):
        self.raw = raw
        return self

    def where(self, *args):
        colName = ''
        data = ''
        compareTag = ''

        if len(args) == 0:
            return self
        elif len(args) == 1:
            if self.whereCon != '':
                self.whereCon = self.whereCon + ' and %s ' % (args[0])
            else:
                self.whereCon = ' where %s ' %  (args[0])
            return self
        elif len(args) == 2:
            compareTag = '='
            colName = args[0]
            data = args[1]
            if '__in' in colName:
                compareTag = 'in'
                colName = colName.replace('__in', '')
                if type(data) == list:
                    itemList = [ "'" + str(item) + "'" for item in data ]
                    data = '(' + ','.join(itemList) + ')'
                elif type(data) == str:
                    splitList = data.split(',')
                    itemList = [ "'" + str(item).strip() + "'" for item in splitList ]
                    data = '(' + ','.join(itemList) + ')'

        elif len(args) == 3:
            colName = args[0]
            compareTag = args[1]
            data = args[2]
        else:
            raise Exception('where param length not support')

        if self.whereCon != '':
            self.whereCon = self.whereCon + ' and `%s` %s' % (colName,
                                                              compareTag)
        else:
            self.whereCon = ' where `%s` %s' % (colName, compareTag)

        if (type(data) == int) or (type(data) == float):
            self.whereCon = self.whereCon + " %s " % data
        else:
            if compareTag == 'in':
                self.whereCon = self.whereCon + " %s " % data
            else:
                data = pymysql.escape_string(data)
                self.whereCon = self.whereCon + " '%s' " % data

        return self

    def orWhere(self, *args):
        colName = ''
        data = ''
        compareTag = ''

        if len(args) == 0:
            return self
        elif len(args) == 1:
            if self.whereCon != '':
                self.whereCon = self.whereCon + ' or %s ' % (args[0])
            else:
                self.whereCon = ' where %s ' %  (args[0])
            return self
        elif len(args) == 2:
            compareTag = '='
            colName = args[0]
            data = args[1]
            if '__in' in colName:
                compareTag = 'in'
                colName = colName.replace('__in', '')
                if type(data) == list:
                    itemList = [ "'" + str(item) + "'" for item in data ]
                    data = '(' + ','.join(itemList) + ')'
                elif type(data) == str:
                    splitList = data.split(',')
                    itemList = [ "'" + str(item).strip() + "'" for item in splitList ]
                    data = '(' + ','.join(itemList) + ')'

        elif len(args) == 3:
            colName = args[0]
            compareTag = args[1]
            data = args[2]
        else:
            raise Exception('where param length not support')

        if self.whereCon != '':
            self.whereCon = self.whereCon + ' or `%s` %s' % (colName,
                                                              compareTag)
        else:
            self.whereCon = ' where `%s` %s' % (colName, compareTag)

        if (type(data) == int) or (type(data) == float):
            self.whereCon = self.whereCon + " %s " % data
        else:
            if compareTag == 'in':
                self.whereCon = self.whereCon + " %s " % data
            else:
                data = pymysql.escape_string(data)
                self.whereCon = self.whereCon + " '%s' " % data

        return self


    def andWhere(self, *args):
        return self.where(*args)

    def getSql(self, cmd):
        #cmd 'insert, delete, update, select'
        if self.raw != '':
            return self.raw
        
        sql = ''
        if cmd == 'select':
            field_str = '*'
            if len(self.fields) == 0:
                field_str = '*'
            else:
                if len(self.fields) == 1:
                    if self.fields[0] == '' or self.fields[0] == '*':
                        field_str = '*'
                    else:
                        field_str = ','.join([ '`' + colname + '`' for colname in self.fields ])
            sql = 'select %s from %s %s' % (field_str, self.tableName, self.whereCon)
        if cmd == 'insert':
            sql_list = []
            if type(self.updateData) == list:
                sql = 'insert into `%s`' % self.tableName
                for row in self.updateData:
                    cols = []
                    for k, v in row.items():
                        cols.append("`" + k + "`")
                    col = ','.join(cols)
                    sql = sql + '(' + col + ') values'
                    break

                values = []
                for row in self.updateData:
                    tup = []
                    for k, v in row.items():
                        if type(v) == int or type(v) == float:
                            tup.append(str(v))
                        elif type(v) == str or type(v) == unicode:
                            tup.append("'" + v + "'")
                    tupStr = '(' + ','.join(tup) + ')'
                    values.append(tupStr)
                valueStr = ', '.join(values)
                sql = sql + valueStr
                self.whereCon = ''
                return sql
            if type(self.updateData) == dict:
                cols = []
                for k, v in self.updateData.items():
                    cols.append("`" + k + "`")
                sql = 'insert into %s' % (self.tableName) + '(' + ','.join(
                    cols) + ') values('
                for k, v in self.updateData.items():
                    if type(v) == int or type(v) == float:
                        sql = sql + str(v) + ','
                    if type(v) == str or type(v) == unicode:
                        sql = sql + "'" + v + "',"
                sql = sql[0:-1]
                sql = sql + ')'
                self.whereCon = ''
                return sql
        if cmd == 'delete':
            sql = """delete from %s %s""" % (self.tableName, self.whereCon)
        if cmd == 'update':
            sql = "update %s set" % (self.tableName)
            for k, v in self.updateData.items():
                if type(v) == int:
                    sql = sql + """ %s=%s,""" % ("`" + k + "`", v)
                if type(v) == str:
                    sql = sql + """ %s='%s',""" % ("`" + k + "`", v)
            sql = sql[0:-1]
            sql = sql + ' ' + self.whereCon
        if cmd == 'count':
            sql = """select count(*) as cnt from %s %s""" % (self.tableName,
                                                             self.whereCon)

        self.whereCon = ''
        return sql

    def data(self, updateData):
        if type(updateData) == list:
            for i in range(len(updateData)):
                for k, v in updateData[i].items():
                    if type(v) == str:
                        updateData[i][k] = pymysql.escape_string(v)
        elif type(updateData) == dict:
            for k, v in updateData.items():
                if type(v) == str:
                    updateData[k] = pymysql.escape_string(v)
        else:
            pass
        self.updateData = updateData
        return self

    def get(self):
        sql = self.getSql('select')
        return self.db.query(sql)

    def values(self, field_name):
        rows = self.get()
        ret = []
        if rows != None:
            for row in rows:
                ret.append(row.get(field_name))
        return ret

    def first(self):
        rows = self.get()
        if len(rows) > 0:
            row = rows[0]
            return row
        return None

    def update(self, retSql=False):
        sql = self.getSql('update')
        return self.db.update(sql)

    def insert(self):
        sql = self.getSql('insert')
        self.db.update(sql)

    def delete(self):
        sql = self.getSql('delete')
        return self.db.update(sql)

    def count(self):
        sql = self.getSql('count')
        rows = self.db.query(sql)
        try:
            row = rows[0]
            num = int(row['cnt'])
        except:
            num = len(rows)
        return num

    def limit(self, limitNum, endpos = None):
        if not str(limitNum).isdigit():
            raise Exception('param limit is not a number')
        if '.' in str(limitNum):
            raise Exception('param limit is not a integer, but a float')
        limitNum = int(limitNum)
        if limitNum < 0: raise Exception('param limit < 0')
        self.whereCon = self.whereCon + " limit %s" % limitNum

        if type(endpos) == int or type(endpos) == str:
            if not str(endpos).isdigit():
                raise Exception('param limit is not a number')
            if int(endpos) < 0 or '.' in str(endpos):
                raise Exception('param limit is invalid')
            self.whereCon = self.whereCon + ", " + str(endpos)

        return self

    def orderBy(self, colName, order=ORDER_ASC):
        if order not in ['asc', 'desc', 'ASC', 'DESC']:
            raise Exception('order by order is invalid')
        self.whereCon = self.whereCon + " order by `%s` %s" % (colName, order)
        return self


class DBCluster(object):
    def __init__(self, masterDb, slaveDbList=[]):
        self._masterDb = masterDb
        self._slaveDbList = slaveDbList
        self._slaveDbCount = len(slaveDbList)
        self._curSlave = 0

    def query(self, sql):
        if len(self._slaveDbList) == 0:
            return self._masterDb.query(sql)
        self._curSlave = self._curSlave + 1
        slaveIndex = self._curSlave % self._slaveDbCount
        slaveDb = self._slaveDbList[slaveIndex]
        return slaveDb.query(sql)

    def queryInstance(self):
        self._curSlave = self._curSlave + 1
        slaveIndex = self._curSlave % self._slaveDbCount
        return slaveIndex

    def update(self, sql):
        return self._masterDb.update(sql)


class DBProxy(object):
    def __init__(self, masterDb, slaveDbList=[]):
        self._masterDb = masterDb
        self._slaveDbList = slaveDbList
        self._slaveDbCount = len(slaveDbList)
        self._curSlave = 0

    def query(self, sql):
        if len(self._slaveDbList) == 0:
            return self._masterDb.query(sql)
        self._curSlave = self._curSlave + 1
        slaveIndex = self._curSlave % self._slaveDbCount
        slaveDb = self._slaveDbList[slaveIndex]
        return slaveDb.query(sql)

    def queryInstance(self):
        self._curSlave = self._curSlave + 1
        slaveIndex = self._curSlave % self._slaveDbCount
        return slaveIndex

    def update(self, sql):
        return self._masterDb.update(sql)


DEFINE_VERSION = '0.0.38'
from setuptools import setup

requireList = [
    'lz4==3.1.3',
    'requests==2.25.1',
    'ujson==1.35',
    'uvloop==0.19.0'
    'uvicorn==0.18.0'
    #'PyMySQL==0.10.1',
    #'DBUtils==1.3',
]

setup(
    name='sunyata',
    version=DEFINE_VERSION,
    description='Light, simple, asynchronous RPC framework for Python',
    author='tank',
    license='MIT',
    platforms="any",
    install_requires=requireList,
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    keywords='sunyata',
    packages=[
        'sunyata', 
        'sunyata/rpc', 
        'sunyata/algorithm', 
        'sunyata/cli',
        'sunyata/http'
    ],
    include_package_data=True,
    entry_points = {
        'console_scripts' : [
            'sunyata=sunyata.cli.entry:main'
        ]
    }
)


import ujson
import requests
from sunyata.util import local_ip
from sunyata.consul import Instance


class EtcdApi(object):

    def __init__(self, host: str, port: int):
        self.host = host
        self.port = port
        self.baseUrl = 'http://%s:%s' % (self.host, self.port)
        self.timeout = 5
        self.headers = {}

    def registService(self, serviceName: str, port = 80, address=None, ttl=30):
        url = self.baseUrl + '/v2/keys/sunyata-services/%s?dir=true&ttl=%s' % (serviceName, ttl)
        r = requests.put(url, timeout=self.timeout, headers=self.headers)
        if address:
            key  = address + ':' + str(port)
        else:
            key = local_ip() + ':' + str(port)
        val = key
        url = self.baseUrl + '/v2/keys/sunyata-services/%s/%s?value=%s&ttl=%s' % (serviceName, key, val, ttl)
        r = requests.put(url, timeout=self.timeout, headers=self.headers)
        print(r.text, r.status_code)

    def getServiceInstanceList(self, serviceName: str) -> list:
        url = self.baseUrl + '/v2/keys/sunyata-services/%s' % serviceName
        r = requests.get(url, timeout=self.timeout, headers=self.headers)
        dic = ujson.loads(r.text)
        nodes = dic.get('node').get('nodes', [])
        instanceList = []
        for node in nodes:
            value = node.get('value')
            ip, port = value.split(':')
            instance = Instance(
                service=serviceName,
                address=ip,
                port=int(port)
            )
            instanceList.append(instance)
        return instanceList
    
    def ttlHeartbeat(self, service, address, port):
        self.registService(serviceName=service, address=address, port=port)


import argparse
import sys
import os
from sunyata.rpc.server import TcpRpcServer

class Cli(object):

    def entryPoint(self):
        args = self.parseArgs()
        self.runServer(args)

    def parseArgs(self):
        parser = argparse.ArgumentParser(description="sunyata console cli tool")
        parser.add_argument('-r', '--run', type=str, help = 'module', required=True, metavar='module', dest='module')
        parser.add_argument('-b', '--bind', type=str, help='bind interface, default 0.0.0.0', default='0.0.0.0', required=False, metavar='0.0.0.0', dest='bind')
        parser.add_argument('-p', '--port', type=int, help='bind port, default 9988', default=9988, required=False, metavar='9988', dest='port')
        args = parser.parse_args()
        try:
            module = args.module
            if module == '':
                raise Exception('')
        except Exception as ex:
            print(ex)
            parser.print_usage()
            sys.exit(1)
        return args

    def runServer(self, args):
        sys.path.append( os.getcwd() )
        __import__(args.module)
        s = TcpRpcServer(args.bind, args.port)
        s.serve()